from ..models import Relation


class RelationAnnotator:
    def annotate(self, text: str, relation: Relation, offset = 0) -> str:
        annotated_text = text[:]

        e1 = relation.e1
        e1_start = e1.location.start - offset
        e1_end = e1.location.end - offset

        e2 = relation.e2
        e2_start = e2.location.start - offset
        e2_end = e2.location.end - offset

        if e1_end < e2_start:
            return annotated_text[:e1_start] + \
                f'<e1>{e1.text}</e1>' + \
                annotated_text[e1_end:e2_start] + \
                f'<e2>{e2.text}</e2>' + \
                annotated_text[e2_end:]

        return annotated_text[:e2_start] + \
            f'<e2>{e2.text}</e2>' + \
            annotated_text[e2_end:e1_start] + \
            f'<e1>{e1.text}</e1>' + \
            annotated_text[e1_end:]
        


from typing import Any, Dict, List, TypeVar

from dataclasses import dataclass, field


NOT_DEFINED_FLAG: int = -1

# pylint: disable=C0103
TLocation = TypeVar('TLocation', bound='Location')
# pylint: enable=C0103

@dataclass(frozen=True)
class Location:
    start: int
    end: int

    @property
    def actual_end(self) -> int:
        return self.end - 1

    def is_in(self: TLocation, other: TLocation) -> bool:
        return self.start >= other.start and self.start <= other.actual_end \
            or self.actual_end >= other.start and self.actual_end <= other.actual_end

    def contains(self: TLocation, other: TLocation) -> bool:
        return other.start >= self.start and other.actual_end <= self.actual_end

    def __repr__(self) -> str:
        return f'({self.start}, {self.end})'

# pylint: disable=C0103
TILocation = TypeVar('TILocation', bound='ILocation')
# pylint: enable=C0103

class ILocation:
    location: Location

    def is_in(self: TILocation, other: TILocation) -> bool:
        return self.location.is_in(other.location)

    def contains(self: TILocation, other: TILocation) -> bool:
        return self.location.contains(other.location)

@dataclass()
class Entity(ILocation):
    label: str
    text: str
    location: Location
    identifier: int = NOT_DEFINED_FLAG
    attributes: Dict[str, Any] = field(default_factory=dict)

    @property
    def start(self) -> int:
        return self.location.start

    @property
    def end(self) -> int:
        return self.location.end

    def __repr__(self) -> str:
        return f'<Entity label="{self.label}" text="{self.text}" span={repr(self.location)}>'

    def __str__(self) -> str:
        return f'##ENTITY_{self.label}_{self.identifier}##'

@dataclass(frozen=True)
class EntityAnnotationResults:
    original_text: str
    annotated_text: str
    entities: List[Entity]

    @property
    def entity_lookup(self):
        return { str(entity): entity for entity in self.entities }

@dataclass(frozen=True)
class Relation:
    label: str
    e1: Entity
    e2: Entity

    @property
    def key(self) -> str:
        return self.create_key(self.e1, self.e2)

    def __repr__(self) -> str:
        return f'<Relation e1="{self.e1.text}" r="{self.label}" e2="{self.e2.text}">'

    @staticmethod
    def create_key(e1: Entity, e2: Entity) -> str:
        return f'{e1.identifier}_{e2.identifier}'

@dataclass(frozen=True)
class Token(ILocation):
    text: str
    location: Location
    order: int
    entities: List[Entity] = field(default_factory=lambda: [])

    def add_entity(self, entity: Entity):
        self.entities.append(entity)

    def __len__(self) -> int:
        return len(self.entities)

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f'<Token text="{self.text}", location={repr(self.location)}, order={self.order}>'

@dataclass(frozen=True)
class TokenGroup(ILocation):
    location: Location
    sentence: str
    tokens: List[Token]



from typing import List, Generator, cast

import re
from ..regexes import RegExLabel
from ..models import EntityAnnotationResults, Relation
from ..utils import flatten


class RelationExtractor:
    def __init__(self, relation_labels: List[RegExLabel]) -> None:
        self._relation_labels = relation_labels

    def extract(self, entity_annotation_results: EntityAnnotationResults) -> List[Relation]:
        entity_lookup = entity_annotation_results.entity_lookup

        def create_relation(label, match: re.Match) -> Relation:
            group = match.groupdict()
            e1_key = group['e1']
            e2_key = group['e2']

            return Relation(
                label,
                entity_lookup[e1_key],
                entity_lookup[e2_key]
            )

        def handler(relationship_label: RegExLabel) -> Generator[Relation, None, None]:
            return (
                create_relation(label, match)
                for label, match in relationship_label.findall(entity_annotation_results.annotated_text)
            )

        return cast(List[Relation], flatten(map(handler, self._relation_labels)))



from .regexes import RegExLabel, RegEx
from .models import Entity, Relation, ILocation, Location, EntityAnnotationResults, Token, TokenGroup
from .entities import EntityExtractor, EntityAnnotator
from .relations import RelationExtractor, RelationAnnotator, RegExRelationLabelBuilder



from typing import List, Generator, Union
from extr import Location
from ..models import Token, TokenGroup


def tokenizer(text: str, sentences: Union[Generator[List[str], None, None], List[List[str]]]) -> Generator[TokenGroup, None, None]:
    offset = 0
    cache = text[:]
    for sentence in sentences:
        counter = 0
        sentence_start = offset
        tokens_in_sentence = []
        for term in sentence:
            start = cache.find(term)
            end = start + len(term)

            actual_term = cache[start:end]
            assert actual_term == term, f'mismatch("{actual_term}", "{term}")'

            tokens_in_sentence.append(
                Token(term, Location(offset + start, offset + end), counter)
            )

            cache = cache[end:]
            offset += end
            counter += 1

        sentence_text = text[sentence_start:offset]

        yield TokenGroup(
            Location(sentence_start, offset),
            sentence_text,
            tokens_in_sentence
        )



from typing import List, cast

from ..regexes import RegExLabel
from ..utils import flatten
from ..models import Location, Entity


class EntityExtractor:
    def __init__(self, regex_labels: List[RegExLabel]) -> None:
        self._regex_labels = regex_labels

    def get_entities(self, text: str) -> List[Entity]:
        def handler(regex_label: RegExLabel):
            return (
                Entity(label, match.group(), Location(*match.span()))
                for label, match in regex_label.findall(text)
            )

        ## sort descending
        all_found_entities = sorted(
            cast(List[Entity], flatten(map(handler, self._regex_labels))),
            key=lambda entity: (entity.end, -entity.start),
            reverse=True
        )

        if len(all_found_entities) == 0:
            return []

        entities = all_found_entities[:1]
        for curr_entity in all_found_entities[1:]:
            prev_entity = entities[-1]

            if curr_entity.is_in(prev_entity):
                continue

            entities.append(curr_entity)

        return entities



from typing import List, TypeVar

from ..regexes import RegEx, RegExLabel


# pylint: disable=C0103
TRegExRelationLabelBuilder = TypeVar('TRegExRelationLabelBuilder', bound='RegExRelationLabelBuilder')
# pylint: enable=C0103

class RegExRelationLabelBuilder:
    def __init__(self: TRegExRelationLabelBuilder, label: str) -> None:
        self._label = label
        self._expressions: List[RegEx] = []

    @property
    def label(self: TRegExRelationLabelBuilder) -> str:
        return self._label

    def add_e1_to_e2(self: TRegExRelationLabelBuilder, e1: str, relation_expressions: List[str], e2: str) -> TRegExRelationLabelBuilder:
        self._expressions.append(
            RegEx(
                list(
                    map(
                        lambda expression: r'(?P<e1>##ENTITY_' + e1 + r'_\d+##)' + expression + r'(?P<e2>##ENTITY_' + e2 + r'_\d+##)',
                        relation_expressions
                    )
                )
            )
        )

        return self

    def add_e2_to_e1(self: TRegExRelationLabelBuilder, e2: str, relation_expressions: List[str], e1: str) -> TRegExRelationLabelBuilder:
        self._expressions.append(
            RegEx(
                list(
                    map(
                        lambda expression: r'(?P<e2>##ENTITY_' + e2 + r'_\d+##)' + expression + r'(?P<e1>##ENTITY_' + e1 + r'_\d+##)',
                        relation_expressions
                    )
                )
            )
        )

        return self

    def build(self: TRegExRelationLabelBuilder) -> RegExLabel:
        return RegExLabel(self.label, self._expressions)



from typing import List
import html
import re

from ..models import Entity, EntityAnnotationResults


class EntityAnnotator:
    def display_entity(self, entity: Entity) -> str:
        return str(entity)

    def annotate(self, text: str, entities: List[Entity], offset = 0) -> EntityAnnotationResults:
        def insert_entity(text: str, entity: Entity) -> str:
            start = entity.start - offset
            end = entity.end - offset
            return text[:start] + self.display_entity(entity) + text[end:]

        annotated_text = text[:]
        for identifer, entity in enumerate(entities):
            entity.identifier = identifer + 1
            annotated_text = insert_entity(annotated_text, entity)

        return EntityAnnotationResults(text, annotated_text, entities)

class LabelOnlyEntityAnnotator(EntityAnnotator):
    def display_entity(self, entity: Entity) -> str:
        return f'<{entity.label}>{entity.text}</{entity.label}>'

class HtmlEntityAnnotator(EntityAnnotator):
    def display_entity(self, entity: Entity) -> str:
        key = re.sub(r' ', '-', entity.label)
        return f'<span class="entity lb-{key}">' + \
            f'<span class="label">{entity.label}</span>' + \
            f'{entity.text}' + \
            '</span>'

    def annotate(self, text: str, entities: List[Entity], offset = 0) -> EntityAnnotationResults:
        return super().annotate(html.escape(text), entities, offset)



from typing import List, Tuple, Optional
import re

from ..utils.iterutils import flatten


class RegEx:
    def __init__(self, expressions: List[str], flags: int = 0, skip_if: Optional[List[str]] = None) -> None:
        self._expressions = expressions
        self._flags = flags
        self._skip_if = set([] if skip_if is None else skip_if)

    def findall(self, text: str) -> List[re.Match]:
        def handler(expression: str) -> list:
            observations = []
            for match in re.finditer(expression, text, self._flags):

                should_skip_match = False
                for skip_if in self._skip_if:
                    if re.search(skip_if, match.group()):
                        should_skip_match = True
                        break

                if not should_skip_match:
                    observations.append(match)

            return observations

        return flatten(map(handler, self._expressions))

class RegExLabel:
    def __init__(self, label: str, regexes: List[RegEx]) -> None:
        self._label = label
        self._regexes = regexes

    def findall(self, text: str) -> List[Tuple[str, re.Match]]:
        def handler(regex) -> list:
            observations = regex.findall(text)
            return list(zip([self._label] * len(observations), observations))

        return flatten(map(handler, self._regexes))



from typing import Dict, List, TypeVar, Generic, Callable, Optional
from copy import deepcopy

T = TypeVar('T')

# pylint: disable=C0103
TQuery = TypeVar('TQuery', bound='Query')
# pylint: enable=C0103

class Query(Generic[T]):
    def __init__(self: TQuery, sequence: List[T]):
        self._sequence = deepcopy(sequence)

    def __filter(self: TQuery, filter_method: Callable[[T], bool]) -> List[T]:
        return list(filter(filter_method, self._sequence))

    def filter(self: TQuery, filter_method: Callable[[T], bool]) -> TQuery:
        self._sequence = self.__filter(filter_method)
        return self

    def find(self: TQuery, find_method: Callable[[T], bool]) -> Optional[T]:
        observations = self.__filter(find_method)

        size = len(observations)
        if size > 1:
            raise Exception('Search was not unique.')

        if size == 1:
            return observations[0]

        return None

    def tolist(self: TQuery) -> List[T]:
        return self._sequence

    def todict(self: TQuery, key_method: Callable[[T], str]) -> Dict[str, T]:
        mapping: Dict[str, T] = {}
        for item in self._sequence:
            mapping[key_method(item)] = item

        return mapping
